/****************************************************************************
 ****************************************************************************/

#include <X11/xpm.h>

#include "jwm.h"

int iconSize = 0;

#ifdef USE_ICONS

#include "x.xpm"

typedef struct IconPathNode {
	char *path;
	struct IconPathNode *next;
} IconPathNode;

static const char *iconSuffix = ".xpm";

static IconNode *iconList;
static IconPathNode *iconPaths;
static IconPathNode *iconPathsTail;

static void SetIconSize();

static IconNode *CreateIconFromXImages(XImage *image, XImage *shape);
static IconNode *CreateIcon();
static void DoDestroyIcon(IconNode *icon);
static void ReadNetWMIcon(ClientNode *np);
static IconNode *GetDefaultIcon();
static IconNode *CreateIconFromData(const char *name, char **data);
static IconNode *CreateIconFromFile(char *fileName);
static IconNode *CreateIconFromBinary(const CARD32 *data, int length);

static int IsSymbolic(char ch);
static char *GetSymbolName(const char *str);
static void ReplaceSymbol(char **str, const char *name, const char *value);
static void ExpandIconPath(char **path);

/****************************************************************************
 * Must be initialized before parsing the configuration.
 ****************************************************************************/
void InitializeIcons() {

	iconList = NULL;
	iconPaths = NULL;
	iconPathsTail = NULL;

}

/****************************************************************************
 ****************************************************************************/
void SetIconSize() {

	XIconSize size;

	if(!iconSize) {

		iconSize = trayHeight - 9;
		if(iconSize > titleHeight - 4) {
			iconSize = titleHeight - 4;
		}

		size.min_width = iconSize;
		size.min_height = iconSize;
		size.max_width = iconSize;
		size.max_height = iconSize;
		size.width_inc = iconSize;
		size.height_inc = iconSize;

		XSetIconSizes(display, rootWindow, &size, 1);

	}

}

/****************************************************************************
 ****************************************************************************/
void DestroyIcons() {

	IconPathNode *pn;

	while(iconList) {
		DoDestroyIcon(iconList);
	}

	while(iconPaths) {
		pn = iconPaths->next;
		Release(iconPaths->path);
		Release(iconPaths);
		iconPaths = pn;
	}
	iconPathsTail = NULL;

}

/****************************************************************************
 ****************************************************************************/
int IsSymbolic(char ch) {

	if(ch >= 'A' && ch <= 'Z') {
		return 1;
	} else if(ch >= 'a' && ch <= 'z') {
		return 1;
	} else if(ch >= '0' && ch <= '9') {
		return 1;
	} else if(ch == '_') {
		return 1;
	} else {
		return 0;
	}

}

/****************************************************************************
 ****************************************************************************/
char *GetSymbolName(const char *str) {

	char *temp;
	int stop;

	if(*str == '$') {
		temp = Allocate(2);
		temp[0] = '$';
		temp[1] = 0;
	} else {
		for(stop = 0; IsSymbolic(str[stop]); stop++);
		temp = Allocate(stop + 1);
		memcpy(temp, str, stop);
		temp[stop] = 0;
	}

	return temp;

}

/****************************************************************************
 ****************************************************************************/
void ReplaceSymbol(char **str, const char *name, const char *value) {

	char *temp;
	int strLength;
	int nameLength;
	int valueLength;
	int x;

	Assert(str);
	Assert(name);

	strLength = strlen(*str);
	nameLength = strlen(name) + 1;
	if(value) {
		valueLength = strlen(value);
	} else {
		valueLength = 0;
	}

	if(valueLength > nameLength) {
		temp = Allocate(strLength - nameLength + valueLength + 1);
		strcpy(temp, *str);
		Release(*str);
		*str = temp;
	}

	temp = strstr(*str, name);
	Assert(temp);
	--temp; /* Account for the "$" */

	if(nameLength > valueLength) {

		/* Move left */
		for(x = 0; temp[x]; x++) {
			temp[x] = temp[x + nameLength - valueLength];
		}
		temp[x] = temp[x + nameLength - valueLength];

	} else if(nameLength < valueLength) {

		/* Move right */
		for(x = strlen(temp); x >= 0; x--) {
			temp[x + valueLength - nameLength] = temp[x];
		}

	}


	if(value) {
		memcpy(temp, value, valueLength);
	}

}

/****************************************************************************
 ****************************************************************************/
void ExpandIconPath(char **path) {

	char *name;
	char *value;
	int x;

	Assert(path);

	for(x = 0; (*path)[x]; x++) {

		if((*path)[x] == '$') {
			name = GetSymbolName(*path + x + 1);
			value = getenv(name);
			ReplaceSymbol(path, name, value);
			Release(name);
			if(value) {
				x += strlen(value) - 1;
			}
		}

	}


}

/****************************************************************************
 ****************************************************************************/
void AddIconPath(const char *path) {

	IconPathNode *ip;
	int length;
	int addSep;

	Assert(path);

	length = strlen(path);
	if(path[length - 1] != '/') {
		addSep = 1;
	} else {
		addSep = 0;
	}

	ip = Allocate(sizeof(IconPathNode));
	ip->path = Allocate(length + addSep + 1);
	strcpy(ip->path, path);
	if(addSep) {
		ip->path[length] = '/';
		ip->path[length + 1] = 0;
	}
	ExpandIconPath(&ip->path);
	ip->next = NULL;

	if(iconPathsTail) {
		iconPathsTail->next = ip;
	} else {
		iconPaths = ip;
	}
	iconPathsTail = ip;

}

/****************************************************************************
 ****************************************************************************/
void PutIcon(const IconNode *icon, Drawable d, GC g, int x, int y) {

	if(icon->image != None) {
		if(icon->mask != None) {
			JXSetClipOrigin(display, g, x, y);
			JXSetClipMask(display, g, icon->mask);
		}

		JXCopyArea(display, icon->image, d, g, 0, 0,
			iconSize, iconSize, x, y);

		if(icon->mask != None) {
			JXSetClipMask(display, g, None);
			JXSetClipOrigin(display, g, 0, 0);
		}
	}

}

/****************************************************************************
 ****************************************************************************/
void LoadIcon(ClientNode *np) {

	IconPathNode *ip;
	char *iconName;

	Assert(np);

	SetIconSize();

	DestroyIcon(np->icon);
	np->icon = NULL;

	/* Attempt to read _NET_WM_ICON for an icon */
	ReadNetWMIcon(np);
	if(np->icon) {
		return;
	}

	/* Attempt to find an icon for this program in the icon directory */
	if(np->instanceName) {
		for(ip = iconPaths; ip; ip = ip->next) {

			iconName = Allocate(strlen(np->instanceName)
				+ strlen(ip->path) + strlen(iconSuffix) + 1);
			strcpy(iconName, ip->path);
			strcat(iconName, np->instanceName);
			strcat(iconName, iconSuffix);

			np->icon = CreateIconFromFile(iconName);

			Release(iconName);
			if(np->icon) {
				return;
			}

		}
	}

	/* Load the default icon */
	np->icon = GetDefaultIcon();

}

/****************************************************************************
 ****************************************************************************/
IconNode *LoadNamedIcon(char *name) {

	IconPathNode *ip;
	IconNode *icon;
	char *temp;

	Assert(name);

	SetIconSize();

	if(name[0] == '/') {

		return CreateIconFromFile(name);

	} else {

		for(ip = iconPaths; ip; ip = ip->next) {

			temp = Allocate(strlen(name) + strlen(ip->path) + 1);
			strcpy(temp, ip->path);
			strcat(temp, name);

			icon = CreateIconFromFile(temp);

			Release(temp);

			if(icon) {
				return icon;
			}

		}

		return NULL;

	}

}

/****************************************************************************
 ****************************************************************************/
void ReadNetWMIcon(ClientNode *np) {

	unsigned long count;
	int status;
	unsigned long extra;
	Atom realType;
	int realFormat;
	unsigned char *data;

	status = JXGetWindowProperty(display, np->window, atoms[ATOM_NET_WM_ICON],
		0, 256 * 256 * 4, False, XA_CARDINAL, &realType, &realFormat, &count,
		&extra, &data);

	if(status == Success && data) {
		np->icon = CreateIconFromBinary((CARD32*)data, count);
		JXFree(data);
	}

}


/****************************************************************************
 ****************************************************************************/
IconNode *GetDefaultIcon() {
	return CreateIconFromData("default", x_xpm);
}

/****************************************************************************
 ****************************************************************************/
IconNode *CreateIconFromData(const char *name, char **data) {

	XImage *image, *shape;
	IconNode *result;

	if(!data) {
		return NULL;
	}

	/* Check if this icon has already been loaded */
	for(result = iconList; result; result = result->next) {
		if(result->name && !strcmp(result->name, name)) {
			return result;
		}
	}

	if(XpmCreateImageFromData(display, data, &image,
		&shape, NULL) == XpmSuccess) {

		result = CreateIconFromXImages(image, shape);
		if(result) {
			result->name = Allocate(strlen(name) + 1);
			strcpy(result->name, name);
		}

		return result;

	} else {
		Debug("invalid xpm");
		return NULL;
	}


}

/****************************************************************************
 ****************************************************************************/
IconNode *CreateIconFromFile(char *fileName) {

	XImage *image, *shape;
	IconNode *result;

	if(!fileName) {
		return NULL;
	}

	/* Check if this icon has already been loaded */
	for(result = iconList; result; result = result->next) {
		if(result->name && !strcmp(result->name, fileName)) {
			return result;
		}
	}

	/* Load the icon */
	if(XpmReadFileToImage(display, fileName, &image,
		&shape, NULL) == XpmSuccess) {
		result = CreateIconFromXImages(image, shape);
		if(result) {
			result->name = Allocate(strlen(fileName) + 1);
			strcpy(result->name, fileName);
		}
		return result;
	} else {
		Debug("error reading xpm: %s", fileName);
		return NULL;
	}

}

/****************************************************************************
 ****************************************************************************/
IconNode *CreateIconFromXImages(XImage *image, XImage *shape) {

	double scalex, scaley;
	double srcx, srcy;
	int destx, desty;
	int index;
	GC gc;
	CARD32 *data;
	char *mask;
	int dbyte, dbit;
	int sbyte, sbit;
	IconNode *result;

	Assert(image);

	scalex = (double)image->width / iconSize;
	scaley = (double)image->height / iconSize;

	data = Allocate(iconSize * iconSize * sizeof(CARD32));
	mask = Allocate(iconSize * iconSize);

	srcy = 0.0;
	dbyte = 0;
	dbit = 0;
	for(desty = 0; desty < iconSize; desty++) {

		srcx = 0.0;
		if(dbit) {
			dbit = 0;
			++dbyte;
		}

		for(destx = 0; destx < iconSize; destx++) {

			index = (int)srcy * image->bytes_per_line + (int)srcx;

			if(shape) {

				sbyte = (int)srcy * shape->bytes_per_line;
				sbyte += (int)srcx >> 3;
				sbit = 7 - ((int)srcx & 7);

				if((shape->data[sbyte] >> sbit) & 1) {
					data[desty * iconSize + destx]
						= XGetPixel(image, (int)srcx, (int)srcy);
					mask[dbyte] |= 1 << dbit;
				} else {
					data[desty * iconSize + destx] = black;
					mask[dbyte] &= ~(1 << dbit);
				}

			} else {

				data[desty * iconSize + destx]
					= XGetPixel(image, (int)srcx, (int)srcy);
				mask[dbyte] |= 1 << dbit;

			}

			srcx += scalex;
			++dbit;
			if(dbit >= 8) {
				dbit = 0;
				++dbyte;
			}

		}
		srcy += scaley;
	}

	JXDestroyImage(image);
	if(shape) {
		JXDestroyImage(shape);
	}

	result = CreateIcon();

	/* Create the mask pixmap */
	result->mask = JXCreatePixmapFromBitmapData(display, rootWindow,
		mask, iconSize, iconSize, white, black, 1);
	Release(mask);

	/* Create the image pixmap */
	result->image = JXCreatePixmap(display, rootWindow, iconSize,
		iconSize, rootDepth);

	image = XCreateImage(display, DefaultVisual(display, rootScreen),
		rootDepth, ZPixmap, 0, (char*)data,
		iconSize, iconSize, 32, 0);

	gc = JXCreateGC(display, result->image, 0, NULL);
	JXPutImage(display, result->image, gc, image, 0, 0, 0, 0,
		iconSize, iconSize);
	JXFreeGC(display, gc);

	Release(image->data);
	image->data = NULL;
	JXDestroyImage(image);

	return result;
}

/****************************************************************************
 ****************************************************************************/
IconNode *CreateIconFromBinary(const CARD32 *input, int length) {

	double scalex, scaley;
	CARD32 height, width;
	CARD32 destx, desty;
	CARD32 index;
	double srcx, srcy;
	CARD32 alpha;
	CARD32 *data;
	XColor color;
	IconNode *result;
	XImage *image;
	GC gc;
	char *maskData;
	int bitOffset, byteOffset;

	if(!input) {
		return NULL;
	}

	width = input[0];
	height = input[1];

	if(width * height + 2 > length) {
		Debug("invalid image size: %d x %d + 2 > %d", width, height, length);
		return NULL;
	}

	scalex = (double)width / iconSize;
	scaley = (double)height / iconSize;

	data = Allocate(iconSize * iconSize * sizeof(CARD32));
	maskData = Allocate(iconSize * iconSize);
	byteOffset = 0;
	bitOffset = 0;

	srcy = 0.0;
	for(desty = 0; desty < iconSize; desty++) {
		srcx = 0.0;
		if(bitOffset) {
			bitOffset = 0;
			++byteOffset;
		}
		for(destx = 0; destx < iconSize; destx++) {

			index = 2 + (CARD32)srcy * width + (CARD32)srcx;

			alpha = input[index] >> 24;
			color.red = ((input[index] >> 16) & 0xFF) * 257;
			color.green = ((input[index] >> 8) & 0xFF) * 257;
			color.blue = (input[index] & 0xFF) * 257;

			if(alpha >= 128) {
				GetColor(&color);
				data[desty * iconSize + destx] = color.pixel;
				maskData[byteOffset] |= (1 << bitOffset);
			} else {
				data[desty * iconSize + destx] = black;
				maskData[byteOffset] &= ~(1 << bitOffset);
			}

			++bitOffset;
			if(bitOffset >= 8) {
				bitOffset = 0;
				++byteOffset;
			}

			srcx += scalex;

		}
		srcy += scaley;
	}

	result = CreateIcon();

	result->mask = JXCreatePixmapFromBitmapData(display, rootWindow,
		maskData, iconSize, iconSize, white, black, 1);
	Release(maskData);

	result->image = JXCreatePixmap(display, rootWindow,
		iconSize, iconSize, rootDepth);
	image = XCreateImage(display, DefaultVisual(display, rootScreen),
		rootDepth, ZPixmap, 0, (char*)data, iconSize, iconSize, 32, 0);
	gc = JXCreateGC(display, result->image, 0, NULL);
	JXPutImage(display, result->image, gc, image, 0, 0, 0, 0,
		iconSize, iconSize);
	Release(image->data);
	image->data = NULL;
	JXDestroyImage(image);
	JXFreeGC(display, gc);

	return result;

}

/****************************************************************************
 ****************************************************************************/
IconNode *CreateIcon() {

	IconNode *icon;

	icon = Allocate(sizeof(IconNode));
	icon->name = NULL;
	icon->image = None;
	icon->mask = None;

	if(iconList) {
		iconList->prev = icon;
	}
	icon->next = iconList;
	icon->prev = NULL;
	iconList = icon;

	return icon;

}

/****************************************************************************
 ****************************************************************************/
void DoDestroyIcon(IconNode *icon) {
	if(icon) {
		if(icon->image) {
			JXFreePixmap(display, icon->image);
		}
		if(icon->mask != None) {
			JXFreePixmap(display, icon->mask);
		}
		if(icon->name) {
			Release(icon->name);
		}

		if(icon->prev) {
			icon->prev->next = icon->next;
		} else {
			iconList = icon->next;
		}
		if(icon->next) {
			icon->next->prev = icon->prev;
		}
		Release(icon);
	}
}

/****************************************************************************
 ****************************************************************************/
void DestroyIcon(IconNode *icon) {

	if(icon && !icon->name) {
		DoDestroyIcon(icon);
	}

}

#endif /* USE_ICONS */

